# -*- coding: utf-8 -*-
"""Module with data plugins that represent files of completed calculations jobs that have been stashed."""

# AUTO-GENERATED

# yapf: disable
# pylint: disable=wildcard-import

from .base import *
from .folder import *

__all__ = (
    'RemoteStashData',
    'RemoteStashFolderData',
)

# yapf: enable
